#include "Exemplar.h"
#include "Local.h"
#include "Parallel.h"

IplImage* exmpExtractPatch(IplImage *img, int r, int c, int size) {
    IplImage* ret = cvCreateImage(cvSize((size*2)+1,(size*2)+1),img->depth,img->nChannels); 
    for(int a = -size; a <= size; a++) {
        for(int b = -size; b <= size; b++) {
            int row = r + a, col = c + b;
            if(row >= img->height) row = img->height-1;
            if(row < 0) row = 0;
            if(col >= img->width) col = img->width-1;
            if(col < 0) col = 0;
            int x = row - (r - size), y = col - (c - size);
            cvSet2D(ret, x, y, cvGet2D(img, row, col));
        }
    }
    return ret;
}

void exmpBorderFront(IplImage *msk, vector< pair<int, int> >& borderPixels) {
    if(!msk) return;
    borderPixels.clear();
    for(int a = 0; a < msk->height; a++) {
        for(int b = 0; b < msk->width; b++) {
            if( iPixel(msk, a, b) < 128 )  continue;
            if( iPixel(msk, a+1, b) < 128 || 
                iPixel(msk, a, b+1) < 128 || 
                iPixel(msk, a-1, b) < 128 || 
                iPixel(msk, a, b-1) < 128 ) {
                borderPixels.push_back(make_pair(a, b));
            }
        }
    }
    return;
}

void exmpListSourcePatches(IplImage *msk, vector< pair<int, int> >& sourcePatches, 
                           int patchSize, int stepSize) {
    if(!msk) return;
    sourcePatches.clear();
    bool foundUncertain = false;
    for(int a = patchSize; a < msk->height - patchSize; a+=stepSize) {
        for(int b = patchSize; b < msk->width - patchSize; b+=stepSize) {
            foundUncertain = false;
            for(int a1 = a - patchSize; a1 <= a + patchSize && !foundUncertain; a1++) {
                for(int b1 = b - patchSize; b1 <= b + patchSize && !foundUncertain; b1++) {
                    if( iPixel(msk, a1, b1) < 128 ) {
                        foundUncertain = true;
                        break;
                    }
                }
            }
            if( foundUncertain ) continue;
            sourcePatches.push_back(make_pair(a, b));
        }
        if( processEvents() ) {
            sourcePatches.clear();
            return;
        }
    }
}

double exmpGetPatchDistance( IplImage *img, IplImage *msk, int x1, int y1, int x2, int y2, int patchSize,
                             double earlyExit, double *result) {
    double dis = 0.0;
    for(int a = -patchSize; a <= patchSize; a++) {
        for(int b = -patchSize; b <= patchSize; b++) {
            int cx1 = x1 + a, cy1 = y1 + b;
            int cx2 = x2 + a, cy2 = y2 + b;
            if( cx1 < 0 || cx1 >= img->height ) continue;
            if( cy1 < 0 || cy1 >= img->width ) continue;
            if( cx2 < 0 || cx2 >= img->height ) continue;
            if( cy2 < 0 || cy2 >= img->width ) continue;
            if( iPixel(msk, cx1, cy1) > 128 ) continue;
            if( earlyExit >= 0.0 && dis > earlyExit ) return 1e50;
            for(int k = 0; k < img->nChannels; k++) {
                dis += iSq( iPixel( img, cx1, cy1, k ) - iPixel( img, cx2, cy2, k ) );
            }
        }
    }
    if( result != NULL ) (*result) = dis;
    return dis;
}

void exmpFillPatch( IplImage *img, IplImage *msk, int sx, int sy, int dx, int dy, int patchSize ) {
    for(int a = -patchSize; a <= patchSize; a++) {
        for(int b = -patchSize; b <= patchSize; b++) {
            int cx1 = sx + a, cy1 = sy + b;
            int cx2 = dx + a, cy2 = dy + b;
            if( iPixel( msk, cx2, cy2 ) < 128.0 ) continue;
            if( cx1 < 0 || cy1 < 0 || cx2 < 0 || cy2 < 0 ) continue;
            if( cx1 >= img->height || cy1 >= img->width || 
                cx2 >= img->height || cy2 >= img->width ) continue;
            iSixel( msk, cx2, cy2, 0.0 );
            cvSet2D( img, cx2, cy2, cvGet2D( img, cx1, cy1 ) );
        }
    }
}

double exmpGetConfidence( IplImage *cp, int x, int y, int patchSize ) {
    double thelta = 0;
    for(int a = x - patchSize; a <= x + patchSize; a++) {
        for(int b = y - patchSize; b <= y + patchSize; b++) {
            thelta += iPixel(cp, a, b);
        }
    }
    int patchDiameter = (patchSize*2)+1;
    thelta /= (double)(patchDiameter*patchDiameter);
    if( thelta < 0.0 ) thelta = 0.0;
    if( thelta > 1.0 ) thelta = 1.0;
    return thelta;
}

void exmpUpdateConfidence( IplImage *img, IplImage *msk, IplImage *cp, int x, int y, int patchSize ) {
    if( patchSize <= 0 || !img || !msk || !cp ) return;
    double thelta = exmpGetConfidence( cp, x, y, patchSize );
    for(int a = x - patchSize; a <= x + patchSize; a++) {
        for(int b = y - patchSize; b <= y + patchSize; b++) {
            if( iPixel(msk, a, b) < 128 ) continue;
            iSixel( cp, a, b, thelta );
            if( a == x && b == y ) {
            }
        }
    }
}


Vector2D exmpGradient( IplImage *image, IplImage *mask, int x, int y ) {
    double dx = 0.0, dy = 0.0, d = 0.0;
    int r = 1;
    while( d == 0.0 ) {
        double k = (1.0 / (2.0 * (double)(r)));
        double xpr = iPixel(image, x + r, y, -1);
        double xmr = iPixel(image, x - r, y, -1);
        double ypr = iPixel(image, x, y + r, -1);
        double ymr = iPixel(image, x, y - r, -1);
        double cdx = xpr - xmr;
        double cdy = ypr - ymr;
        if( iPixel(mask, x + r, y) > 128 || iPixel(mask, x - r, y) > 128 ) cdx = 0.0;
        if( iPixel(mask, x, y + r) > 128 || iPixel(mask, x, y - r) > 128 ) cdy = 0.0;
        dx = k * cdx; dy = k * cdy;
        d = sqrt(iSq(dx)+iSq(dy));
        r++;
        if( r > EXMP_GRADIENT_MAX_RADIUS ) return Vector2D(0.0, 0.0);
    }
    return Vector2D( dx , dy );
}

double exmpCalculateDatapoint( IplImage *img, IplImage *msk, int x, int y, 
                          int windowSize, int patchSize ) {
    if( patchSize <= 0 || !img || !msk ) return 0.0;

    // Create image patches
    IplImage *mskWindow = exmpExtractPatch( msk, x, y, windowSize );
    IplImage *mskdx = cvCreateImage( cvGetSize(mskWindow), IPL_DEPTH_32F, 1 );
    IplImage *mskdy = cvCreateImage( cvGetSize(mskWindow), IPL_DEPTH_32F, 1 );
    
    //Sobel differentiate mask image
    cvSobel( mskWindow, mskdx, 0, 1, 3 );
    cvSobel( mskWindow, mskdy, 1, 0, 3 );
    
    // Setup for isotope
    Vector2D newIsotope(0.0, 0.0);
    Vector2D maxIsotope(0.0, 0.0);
    double maxIsotopeLength = 0.0;
    
    // Loop through to find maximum isotopes
    for(int a = -patchSize; a <= patchSize; a++) {
        for(int b = -patchSize; b <= patchSize; b++) {
            if( iPixel( msk, x + a, y + b ) > 128 ) continue;
            newIsotope = exmpGradient( img, msk, x + a, y + b );
            double newLength = newIsotope.length();
            if( newLength > maxIsotopeLength ) {
                maxIsotopeLength = newLength;
                maxIsotope = newIsotope;
            }
        }
    }
    // Turn vector 90 degrees
    Vector2D isotope = Vector2D( -maxIsotope.y, maxIsotope.x );
    
    // Get normal and normalize it
    Vector2D normal( iPixel( mskdx, (mskdx->height/2)+1, (mskdx->width/2)+1 ), 
                     iPixel( mskdy, (mskdy->height/2)+1, (mskdy->width/2)+1 ) );
    normal.normalize();
    
    double dataPointValue = fabs(isotope * normal) / EXMP_DATAPOINT_ALPHA;
    if( dataPointValue < 0.0 ) dataPointValue = 0.0;
    
    cvReleaseImage( &mskWindow );
    cvReleaseImage( &mskdx );
    cvReleaseImage( &mskdy );
    
    return dataPointValue;
}

double exmpCalculatePriority( IplImage *img, IplImage *msk, IplImage *cp, int x, int y, int patchSize,
                              double *result) {
    if( patchSize <= 0 || !img || !msk || !cp) return 0.0;
    double ret = exmpGetConfidence(cp, x, y, patchSize) * 
        exmpCalculateDatapoint( img, msk, x, y, 
        EXMP_DATAPOINT_NORMAL_WINDOWSIZE, patchSize );
    if( result ) (*result) = ret;
    return ret;
}

void exmpInitializePriority( IplImage *img, IplImage *msk, IplImage *cp ) {
    if( !img || !msk || !cp ) return;
    for(int a = 0; a < msk->height; a++) {
        for(int b = 0; b < msk->width; b++) {
            if( iPixel(msk, a, b) < 128 ) iSixel( cp, a, b, 1.0 );
            else iSixel( cp, a, b, 0.0 );
        }
    }
}

void exmpSourceMask( IplImage *imsk, IplImage *smsk, int dis ) {
    cvDilate(smsk, smsk, 0, dis);
    for(int a = 0; a < smsk->height; a++) {
        for(int b = 0; b < smsk->width; b++) {
            if( iPixel(imsk, a, b) > 128 ) {
                iSixel(smsk, a, b, 0.0);
            }
        }
    }
}

void exmpInPaint(IplImage *img, IplImage *msk, int iter, int patchSize, int boundary, int skip, int method, bool GPU) {
    if(!img || !msk || iter<=0 || patchSize<=1) return;

    IplImage *iterImg = iCloneFloat( img, 1.0 );
    IplImage *iterMask = cvCloneImage( msk );
    IplImage *sourceMask = cvCloneImage( msk );
    IplImage *confidence = iCloneFloat( msk, 1.0 );

    exmpInitializePriority( iterImg, iterMask, confidence );

    vector< pair<int, int> > borderPixels;
    vector< pair<int, int> > sourcePatches;
    vector<double> borderPriority;

    exmpSourceMask( iterMask, sourceMask, boundary );
    exmpListSourcePatches( sourceMask, sourcePatches, patchSize, skip );
    if( sourcePatches.size() <= 0 ) {
        cvReleaseImage( &iterImg );
        cvReleaseImage( &iterMask );
        cvReleaseImage( &sourceMask );
        cvReleaseImage( &confidence );
        return;
    }

    for(int i = 0; i < iter; i++) {
        if( processEvents() ) { break; }
        // Determine border front pixels
        exmpBorderFront( iterMask, borderPixels );
        if( borderPixels.size() <= 0 ) break;

        // Calculate priorities and find first priority pixel
        double pr = 0.0, maxpr = -1.0;
        int px = -1, py = -1;
        printl("Calculating border pixel priorities...\n");
        borderPriority.clear();
        for(int a = 0; a < borderPixels.size(); a++) {
            int x = borderPixels[a].first;
            int y = borderPixels[a].second;
            exmpCalculatePriority( iterImg, iterMask, confidence, x, y, patchSize, &pr );
            if( method != EXMP_METHOD ) borderPriority.push_back( pr );
            if( pr > maxpr ) {
                maxpr = pr;
                px = x;
                py = y;
            }
        }
        // If we're out of confidence, pick a random
        if( maxpr <= 0.00000001 ) {
            int r = rand()%borderPixels.size();
            px = borderPixels[r].first;
            py = borderPixels[r].second;
        }
        if( px < 0 || py < 0 ) break;

        // Find exemplar image patch in source image
        int sx = -1, sy = -1;
        if( method == EXMP_METHOD ) {
            double dist = 0.0, mindist = -1.0;
            if( gpuEnabled && GPU ) {
                #ifdef _GPU_
                    printl("Calculating exemplar (GPU) ...\n");
                    float *result = gpuComputeDistanceList( iterImg, iterMask, patchSize, sourcePatches, px, py );
                    if( !result ) {
                        printl("Error: Result NULL!\n");
                    } else {
                        for(int a = 0; a < sourcePatches.size(); a++) {
                            if( result[a] < mindist || mindist < -0.5 ) {
                                mindist = result[a];
                                sx = sourcePatches[a].first;
                                sy = sourcePatches[a].second;
                            }
                        }
                        delete[] result;
                   }
                #endif
            } else {
                printl("Calculating exemplar...\n");
                dist = 0.0, mindist = -1.0;
                for(int a = 0; a < sourcePatches.size(); a++) {
                    int x = sourcePatches[a].first;
                    int y = sourcePatches[a].second;
                    exmpGetPatchDistance( iterImg, iterMask, px, py, x, y, patchSize, mindist, &dist );
                    if( dist < mindist || mindist < -0.5) {
                        mindist = dist;
                        sx = x;
                        sy = y;
                    }
                }
            }
        } else {
            localGetExemplarPatch( iterImg, iterMask, confidence, px, py, patchSize, 
                sourcePatches, borderPriority, &sx, &sy );
        }
        if( sx < 0 || sy < 0 ) break;

        // Update confidence values
        printl("Updating confidence...\n");
        exmpUpdateConfidence( iterImg, iterMask, confidence, px, py, patchSize );
        
        // Fill in the gaps in the target patch using the exemplar
        printl("Filling in information...\n");
        exmpFillPatch( iterImg, iterMask, sx, sy, px, py, patchSize );
        printl("Confidence %lf\n", iPixel(confidence, px, py) );

        cvCvtScale( iterImg, img, 1.0 );
        windowUpdate();

    }

    cvCvtScale( iterImg, img, 1.0 );
    cvCvtScale( iterMask, msk, 1.0 );
    cvThreshold(msk, msk, 128.0, 255.0, CV_THRESH_BINARY);


    cvReleaseImage( &iterImg );
    cvReleaseImage( &iterMask );
    cvReleaseImage( &sourceMask );
    cvReleaseImage( &confidence );
}