#include "Inpaint.h"
#include "Vector.h"
#include "Diffusion.h"
#include "Toolbar.h"
#include "Contour.h"
#include "Fast.h"
#include "Exemplar.h"

// =======> Inpainter Methods

Inpainter::Inpainter() {
    imageSource = NULL;
    imageMask = NULL;
    for(int a = 0; a < NUM_IMAGE_SEPARATION; a++) image[a] = NULL;
    for(int a = 0; a < NUM_IMAGE_SEPARATION; a++) imageResult[a] = NULL;
    imageResultCombined = NULL;
    methodInpaint = 0;
    methodSeparation = 0;

    pGaussianOctaves = 8;
    pFDIIKernel = 1;
    pFDIIIterations = 5;
    pBertalmioInpaintSteps = 15;
    pBertalmioDiffuseSteps = 4;
    pBertalmioClimbRate = 0.02;
    pBertalmioClimbDecay = 0.98922;
    pTeleaRadius = 3;
    pNavierRadius = 3;
    pExmpIterations = 1000;
    pExmpPatchSize = 4;
    pExmpBoundary = 10;
    pExmpGPU = false;
}

Inpainter::~Inpainter() { reset(); }

bool Inpainter::RGBCapableMethod() {
    return ( this->methodInpaint == INP_TELEA ||
             this->methodInpaint == INP_NAVIER ||
             this->methodInpaint == INP_EXEMPLAR ||
             this->methodInpaint == INP_LOCAL );
}

void Inpainter::reset() {
    if(imageSource) cvReleaseImage(&imageSource);
    if(imageMask) cvReleaseImage(&imageMask);
    if(imageResultCombined) cvReleaseImage(&imageResultCombined);
    for(int a = 0; a < NUM_IMAGE_SEPARATION; a++) {
        if(image[a]) cvReleaseImage(&image[a]);
        if(imageResult[a]) cvReleaseImage(&imageResult[a]);
    }
}

void Inpainter::load(const char* fileName, const char* maskName) {
    this->reset();
    char buffer[BUFFER_LEN];
    if(!fileName) return;

    imageSource = cvLoadImage(fileName, true);
    if(!imageSource) return;
    if( maskName[0] != '\0' ) {
        imageMask = cvLoadImage(maskName, false);
        cvThreshold(imageMask, imageMask, 128.0, 255.0, CV_THRESH_BINARY);
    }
    if( !imageMask ) {
        printl("Mask loading failed...using blank mask.");
        imageMask = cvCreateImage(cvGetSize(imageSource), IPL_DEPTH_8U, 1);
        iZero( imageMask );
    } else {
        if( imageMask->height != imageSource->height ||
            imageMask->width != imageSource->width ) {
            printl("Mask is not same size as image!");
            printl("Using blank mask.");
            imageMask = cvCreateImage(cvGetSize(imageSource), IPL_DEPTH_8U, 1);
            iZero( imageMask );
        }
    }
    for(int a = 0; a < NUM_IMAGE_SEPARATION; a++) {
        image[a] = cvCreateImage(cvGetSize(imageSource), IPL_DEPTH_32F, 1);
        imageResult[a] = cvCreateImage(cvGetSize(imageSource), IPL_DEPTH_32F, 1);
    }
    imageResultCombined = cvCreateImage(cvGetSize(imageSource), IPL_DEPTH_8U, 3);
    return;
}

void Inpainter::separate() {
    if(!imageSource) return;
    double pixelB, pixelG, pixelR, pixel;
    double tempPixel, tempLength;
    for(int a = 0; a < imageSource->height; a++) {
        for(int b = 0; b < imageSource->width; b++) {
            pixelB = iPixel( imageSource, a, b, 0 );
            pixelG = iPixel( imageSource, a, b, 1 );
            pixelR = iPixel( imageSource, a, b, 2 );
            switch( this->methodSeparation ) {
                case SEP_GREYSCALE:
                    pixel =(((double)(pixelB)) * GREYSCALE_LUMIN_B +
                            ((double)(pixelG)) * GREYSCALE_LUMIN_G +
                            ((double)(pixelR)) * GREYSCALE_LUMIN_R);
                    for(int c = 0; c < NUM_IMAGE_SEPARATION; c++)
                        iSixel( image[c], a, b, pixel );
                    break;
                case SEP_DESATURATE:
                    pixel =(((double)(pixelB)) * DESATURATE_SCALE +
                            ((double)(pixelG)) * DESATURATE_SCALE +
                            ((double)(pixelR)) * DESATURATE_SCALE);
                    for(int c = 0; c < NUM_IMAGE_SEPARATION; c++)
                        iSixel( image[c], a, b, pixel );
                    break;
                case SEP_RGB:
                    iSixel( image[0], a, b, pixelB );
                    iSixel( image[1], a, b, pixelG );
                    iSixel( image[2], a, b, pixelR );
                    break;
                case SEP_VECTOR:
                    // Calculate vector length
                    tempPixel = sqrt( iSq((double)pixelB) + iSq((double)pixelG) + iSq((double)pixelR) );
                    pixel = ( tempPixel * VECTOR_LENGTH_SCALE );
                    iSixel( image[0], a, b, pixel );
                    // Horizontal Angle
                    Vector2D pixelVector((double)pixelB, (double)pixelR);
                    tempLength = pixelVector.length();
                    pixelVector.normalize();
                    iSixel( image[1], a, b, ( pixelVector.y * VECTOR_SIN_SCALE ) );
                    // Vertical Angle
                    Vector2D pixelVector2(tempLength, (double)pixelG);
                    pixelVector2.normalize();
                    iSixel( image[2], a, b, ( pixelVector2.y * VECTOR_SIN_SCALE ) );
                    break;
            }
        }
    }
}

void Inpainter::prepare() {
    if(!imageSource || !imageResultCombined || !image[0]) return;
    if( RGBCapableMethod() ) {
        cvCopyImage( imageSource, imageResultCombined );
    }
    for(int a = 0; a < NUM_IMAGE_SEPARATION; a++) {
        cvCopyImage( image[a], imageResult[a] );
    }
    iGlobalSaveCount = 0;
}

void Inpainter::inpaint() {
    if(!imageSource || !image[0] || !imageResult[0]) return;
    int nResultProcess = 3;
    if( this->methodSeparation == SEP_GREYSCALE ||
        this->methodSeparation == SEP_DESATURATE ) nResultProcess = 1;
    switch( this->methodInpaint ) {
        case INP_NONE: break;
        case INP_GAUSSIANBLUR:
            for(int a = 0; a < nResultProcess; a++)
                dPyramidGaussianBlur(imageResult[a], imageMask, pGaussianOctaves);
            break;
        case INP_FDII:
            for(int a = 0; a < nResultProcess; a++)
                FDIInpaint( imageResult[a], imageMask, pFDIIKernel, globalInpainter.pFDIIIterations );
            break;
        case INP_BERTALMIO:
            if( pBertalmioClimbRate > 10000.0 ) pBertalmioClimbRate = 10000.0;
            if( pBertalmioClimbRate < 0.0 ) pBertalmioClimbRate = 0.0;
            if( pBertalmioClimbDecay < 0.5 ) pBertalmioClimbDecay = 0.5;
            if( pBertalmioClimbDecay > 1.5 ) pBertalmioClimbDecay = 1.5;
            for(int a = 0; a < nResultProcess; a++) {
                if(processEvents()) return;
                printl("Channel: %d - Applying %d steps of Aniso. Diffusion", a, pBertalmioDiffuseSteps);
                dAnisotripicDiffusion( imageResult[a], imageMask, pBertalmioDiffuseSteps );
                printl("Channel: %d - Applying %d steps of Inpainting", a, pBertalmioDiffuseSteps);
                contourInpaint(imageResult[a], imageMask, imageResult[a], pBertalmioClimbRate,
                               pBertalmioInpaintSteps);
                pBertalmioClimbRate *= pBertalmioClimbDecay;
            }
            break;
        case INP_TELEA:
            cvInpaint( imageSource, imageMask, imageResultCombined, pTeleaRadius, CV_INPAINT_TELEA);
            break;
        case INP_NAVIER:
            cvInpaint( imageSource, imageMask, imageResultCombined, pNavierRadius, CV_INPAINT_NS);
            break;
        case INP_EXEMPLAR:
        case INP_LOCAL:
            exmpInPaint( imageResultCombined, imageMask, pExmpIterations, pExmpPatchSize,
                pExmpBoundary, 1, methodInpaint == INP_EXEMPLAR ? EXMP_METHOD : LOCAL_METHOD, pExmpGPU );
            break;
    }
}

void Inpainter::combine() {
    if(!imageSource || !image[0] || !imageResult[0]) return;
    if( RGBCapableMethod() ) return;
    double pixelB, pixelG, pixelR, k, len;
    for(int a = 0; a < imageResultCombined->height; a++) {
        for(int b = 0; b < imageResultCombined->width; b++) {
            double pixel0 = iPixel( imageResult[0], a, b );
            double pixel1 = iPixel( imageResult[1], a, b );
            double pixel2 = iPixel( imageResult[2], a, b );
            switch( this->methodSeparation ) {
                case SEP_GREYSCALE:
                case SEP_DESATURATE:
                    iSixel( imageResultCombined, a, b, pixel0 );
                    break;
                case SEP_RGB:
                    cvSet2D( imageResultCombined, a, b, cvScalar( pixel0, pixel1, pixel2, 1.0 ) );
                    break;
                case SEP_VECTOR:
                    pixel1 /= VECTOR_SIN_SCALE;
                    pixel2 /= VECTOR_SIN_SCALE;
                    pixel0 /= VECTOR_LENGTH_SCALE;
                    pixelR = pixel1*cos(asin(pixel2));
                    pixelB = cos(asin(pixel1))*cos(asin(pixel2));
                    pixelG = pixel2;
                    if( pixelB < 0 ) pixelB = 0;
                    if( pixelG < 0 ) pixelG = 0;
                    if( pixelR < 0 ) pixelR = 0;
                    pixelB *= pixel0;
                    pixelG *= pixel0;
                    pixelR *= pixel0;
                    cvSet2D( imageResultCombined, a, b, cvScalar( pixelB, pixelG, pixelR, 1.0 ) );
                    break;
            }
        }
    }
}

IplImage* Inpainter::operator [] (int index) {
    if(index < -3 || index >= NUM_IMAGE_SEPARATION) return NULL;
    if(index == -3) return imageResultCombined;
    if(index == -2) return imageMask;
    if(index == -1) return imageSource;
    return image[index];
}