#include "Local.h"
#include "Parallel.h"

bool localTestBorder(IplImage *msk, int a, int b) {
    if(!msk) return false;
    if( iPixel(msk, a, b) < 128 ) return false;
    if( iPixel(msk, a+1, b) < 128 || 
        iPixel(msk, a, b+1) < 128 || 
        iPixel(msk, a-1, b) < 128 || 
        iPixel(msk, a, b-1) < 128 ) {
        return true;
    }
    return false;
}


void localExtractSourcePatchList(IplImage *img, IplImage *msk, int nx, int ny, int patchSize,
                                 const vector< pair< int, int > >& sourcePatches,
                                 vector< pair< int, int > >& output) {
    if( !img || !msk ) return;
    double dist = 0.0;
    vector< pair< double, pair< int, int > > > resultList; 
    for(int a = 0; a < sourcePatches.size(); a++) {
        int x = sourcePatches[a].first;
        int y = sourcePatches[a].second;
        exmpGetPatchDistance( img, msk, nx, ny, x, y, patchSize, -1, &dist );
        resultList.push_back( make_pair( dist, make_pair( x, y ) ) );
    }
    sort( resultList.begin(), resultList.end() );
    for(int a = 0; a < LOCAL_TOP_NEIGHBOUR_PATCHES; a++) {
        if( a >= resultList.size() ) break;
        output.push_back( resultList[a].second );
    }
}

void localGetExemplarPatch(IplImage *img, IplImage *msk, IplImage *cp, int px, int py, int patchSize,
                           const vector< pair< int, int > >& sourcePatches,
                           const vector< double >& borderPriority,
                           int *destx, int *desty, bool GPU) {
    if( !img || !msk ) return;
    // Extract list of neighbour patches
    int nNeighbours = 0;
    vector< pair< int, int > > nPatchList[LOCAL_MAX_NEIGHBOURHOOD_SIZE];
    int neighbourXPos[LOCAL_MAX_NEIGHBOURHOOD_SIZE];
    int neighbourYPos[LOCAL_MAX_NEIGHBOURHOOD_SIZE];
    printl("Extracting source patchlist...\n");
    for(int a = px - LOCAL_TOP_NEIGHBOURHOOD_SIZE; a <= px + LOCAL_TOP_NEIGHBOURHOOD_SIZE; a++) {
        for(int b = py - LOCAL_TOP_NEIGHBOURHOOD_SIZE; b <= py + LOCAL_TOP_NEIGHBOURHOOD_SIZE; b++) {
            if( a < 0 || a >= img->height || b < 0 || b >= img->width ) continue;
            if( a == px && b == py ) continue;
            if( !localTestBorder( msk, a, b ) ) continue;
            localExtractSourcePatchList( img, msk, a, b, patchSize, sourcePatches, nPatchList[nNeighbours] );
            neighbourXPos[nNeighbours] = a;
            neighbourYPos[nNeighbours] = b;
            nNeighbours++;
        }
    }
    // Go through and calculate & minimise joint distribution
    double *distArray = new double[sourcePatches.size() + 1];
    double *psiexp = new double[sourcePatches.size() + 1];
    if( gpuEnabled && GPU ) {
        #ifdef _GPU_
            printl("Calculating similarity term (GPU)...\n");
            float *result = gpuComputeDistanceList( img, msk, patchSize, sourcePatches, px, py );
            if( !result ) {
                printl("Error: Result NULL!\n");
            } else {
                for(int a = 0; a < sourcePatches.size(); a++) {
                    distArray[a] = (double)result[a];
                    distArray[a] *= LOCAL_DIST_SCALE;
                    psiexp[a] = exp(-distArray[a]);
                }
                delete[] result;
           }
        #endif
    } else {
        printl("Calculating similarity term ...\n");
        for(int a = 0; a < sourcePatches.size(); a++) {
            // First term, similarity of image to its target
            int x = sourcePatches[a].first;
            int y = sourcePatches[a].second;
                exmpGetPatchDistance( img, msk, px, py, x, y, patchSize, -1.0, &distArray[a] );
                distArray[a] *= LOCAL_DIST_SCALE;
                psiexp[a] = exp(-distArray[a]);
        }
    }

    vector< pair< double, pair< int, int > > > sourcePatches2;
    vector< pair< int, int > > sourcePatches3;
    for(int a = 0; a < sourcePatches.size(); a++) {
        sourcePatches2.push_back(make_pair(psiexp[a], sourcePatches[a]));
    }
    sort(sourcePatches2.begin(), sourcePatches2.end());
    if( sourcePatches2.size() == 0 ) return;
    int nSourcePatches = 0;
    for(int a = int( sourcePatches2.size() - 1 ); a >= 0; a--) {
        if( nSourcePatches++ > LOCAL_TOP_NEIGHBOUR_PATCHES ) break;
        sourcePatches3.push_back(sourcePatches2[a].second);
    }


    double bestSourcePatch = -1.0;
    int bestX = -1;
    int bestY = -1;
    for(int a = 0; a < sourcePatches3.size(); a++) {
        // Second term, similarity of image to its neighbours
        double xpxpn = -1.0;
        double dist = 0.0;
        double pxpxpn = 1.0;
        int x = sourcePatches3[a].first;
        int y = sourcePatches3[a].second;
        printl("Calculating weighted neighbour term ...\n");
        for(int b = 0; b < nNeighbours; b++) {
            if( processEvents() ) { break; }
            // Calculate similarity to this neighbour
            printf("Neighbour %d / %d: %d / %d - %d patches\n", a, sourcePatches3.size(), b, nNeighbours, nPatchList[b].size());
            xpxpn = -1.0;
            if( gpuEnabled && GPU ) {
                #ifdef _GPU_
                    float *result = gpuComputeDistanceList( img, msk, patchSize, nPatchList[b], x, y );
                    if( !result ) {
                        printl("Error: Result NULL!\n");
                    } else {
                        for(int c = 0; c < nPatchList[b].size(); c++) {
                            if( (double)result[c] < xpxpn || xpxpn < -0.5) {
                                xpxpn = (double)result[c];
                            }
                        }
                        delete[] result;
                   }
                #endif
            } else {
                for(int c = 0; c < nPatchList[b].size(); c++) {
                    int nx = nPatchList[b][c].first;
                    int ny = nPatchList[b][c].second;
                    exmpGetPatchDistance( img, msk, x, y, nx, ny, patchSize, xpxpn, &dist );
                    if( dist < xpxpn || xpxpn < -0.5) {
                        xpxpn = dist;
                    }
                }
            }
            if( xpxpn < -0.5 ) continue;
            // Third term, priority of this neighbour
            double pxpn = 1.0;
            exmpCalculatePriority( img, msk, cp, neighbourXPos[b], neighbourYPos[b], patchSize, &pxpn );
            xpxpn *= LOCAL_DIST_SCALE;
            pxpxpn += ( exp(-xpxpn) * pxpn );
        }
        double P = psiexp[a] * pxpxpn;
        if( P > bestSourcePatch ) {
            bestSourcePatch = P;
            bestX = x;
            bestY = y;
        }
    }

    if(destx) (*destx) = bestX;
    if(desty) (*desty) = bestY;

    delete[] distArray;
    delete[] psiexp;
}